--[[
Name: TreatOrTreat
Revision: Beta-2.0.1
Author(s): falconindy
Description: Warnings for targets and recipient rogues of Tricks of the Trade.
Inspired by: SimpleMD and Misdrection Helper
	]]--

--[[ Addon Declaration/Namespace ]]--
TrickOrTreat = LibStub("AceAddon-3.0"):NewAddon("TrickOrTreat", "AceEvent-3.0", "AceConsole-3.0", "LibSink-2.0")
local addon = TrickOrTreat
local AceConfigDialog = LibStub("AceConfigDialog-3.0")
local AceConfig = LibStub("AceConfig-3.0")
local callbacks = LibStub("CallbackHandler-1.0"):New(TrickOrTreat)

addon.ClassColor = {
	["DEATHKNIGHT"]		= "|cffc41f3b",
	["DRUID"]			= "|cffff7d0a",
	["HUNTER"]			= "|cffabd473",
	["MAGE"]			= "|cff69ccf0",
	["PALADIN"]			= "|cfff58cba",
	["PRIEST"]			= "|cffffffff",
	["ROGUE"]			= "|cfffff569",
	["SHAMAN"]			= "|cff2459ff",
	["WARLOCK"]			= "|cff9482c9",
	["WARRIOR"]			= "|cffc79c6e",
}

local UnitDetailedThreatSituation = _G.UnitDetailedThreatSituation
local UnitExists = _G.UnitExists
local mfloor = math.floor
local tinsert = tinsert
local pairs = pairs
local ipairs = ipairs

local damageMod = 1.15
local pName = UnitName("player")

local function triggerEvent(...)
	callbacks:Fire(...)
end

--[[ Addon Methods ]]--
function addon:OnInitialize()
	self:SetSinkStorage({ ["sink20OutputSink"] = "RaidWarning" })

	self.db = LibStub("AceDB-3.0"):New("ToTDB", self.defaults, "Default")

	self:RegisterChatCommand("tot", function() AceConfigDialog:Open("TrickOrTreat") end )
	self:RegisterChatCommand("trickortreat", function() AceConfigDialog:Open("TrickOrTreat") end )
	
	local optFrame = AceConfigDialog:AddToBlizOptions("TrickOrTreat", "TrickOrTreat")
	self.options.args.profile = LibStub("AceDBOptions-3.0"):GetOptionsTable(self.db)
	AceConfig:RegisterOptionsTable("TrickOrTreat", self.options)
end

function addon:OnEnable()
	self:RegisterEvent("PLAYER_ENTERING_WORLD", "CheckForParty")
	self:RegisterEvent("RAID_ROSTER_UPDATE", "CheckForParty")
	self:RegisterEvent("PARTY_MEMBERS_CHANGED", "CheckForParty")
	self:RegisterEvent("PLAYER_REGEN_DISABLED", function() --[[ self:ClearActors() ]] end) --on entering combat, clear table
	self:RegisterEvent("PLAYER_REGEN_ENABLED", function() --[[self:ReportActors("ALL") ]] end) --on leaving combat, print table
	
	self.dmgCounter_self = 0
	self.dmgCounter_target = 0
	self.extra_dmg = 0
	self.parse_dmg_self = false
	self.parse_dmg_target = false
	self.last_tricked = nil
	
	self.tracking = self.tracking or {}
	
	triggerEvent("NO_TARGET")
end

function addon:CheckForParty(event) --only enable CLEU handler in a group
	--self:Print(event .. ": Checking for party...")
	if (GetNumRaidMembers() > 0) or (GetNumPartyMembers() > 0) then
		--self:Print("Party time!")
		self:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
		triggerEvent("NO_TARGET")
	else
		--self:Print("No more party =(")
		self:UnregisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
		self.last_tricked = nil
		--triggerEvent("NO_PARTY")
	end
end

--[[ Actor Tracking 
tracking = {	--table layout
	{
		name = <actor name>,
		n = <num tricks>,
		dmg = <total dmg>,
	},
	...
}
]]--
do
	local function getActorID(actor)
		if #addon.tracking == 0 then return 0 end
		for k,v in ipairs(addon.tracking) do
			if v.name == actor then
				return k
			end
		end
		return 0
	end

	function addon:AddActor(actor, damage)
		id = getActorID(actor)
		if id > 0 then
			self.tracking[id].n = self.tracking[id].n + 1
			self.tracking[id].dmg = self.tracking[id].dmg + damage
		else
			tinsert(self.tracking, {name = actor, n = 1, dmg = damage})
		end
	end

	function addon:ClearActors()
		if #self.tracking == 0 then return end
		self:Print("Resetting saved actor data.")
		for k in ipairs(self.tracking) do
			self.tracking[k] = nil
		end
	end

	function addon:ReportActors(method, actor)
		if #self.tracking == 0 then return end
		if method == "ALL" then
			self:Print("Extra damage by actor:")
			for _,v in pairs(self.tracking) do
				print(v.name, ":", v.dmg, "(" .. v.n .. ")")
			end
		elseif method == "TARGET" then
			SendChatMessage("Tricks has granted you " .. self.tracking[actor].dmg .. " extra damage.", "WHISPER", nil, actor)
		end
	end
end

do
	local function AddDamage(counter, damage, overkill)
		return counter + damage - overkill
	end
	
	local function ClearCounters()
		addon.dmgCounter_self = 0
		addon.dmgCounter_target = 0
	end
	
	function addon:COMBAT_LOG_EVENT_UNFILTERED(event, timestamp, eventType, srcGUID, sourceName, srcFlags, destGUID, destName, destFlags, spellID, spellName, ...)
		if spellID == 57934 and sourceName == pName then --primer applied
			if eventType == "SPELL_CAST_SUCCESS" then
				self.last_tricked = destName
				self.parse_dmg_self = true
				self:Pour("|cffff6600[Trick or Treat] " .. addon.ClassColor[select(2, UnitClass(destName))] .. destName .. "|r")
				triggerEvent("TRICKS_APPLIED")
				ClearCounters()
			elseif eventType == "SPELL_AURA_REMOVED" and self.dmgCounter_self == 0 then
				triggerEvent("NO_TARGET")
			end
		elseif spellID == 57933 then --transferring threat
			if eventType == "SPELL_AURA_APPLIED" and sourceName == pName then
				self.parse_dmg_target = true
				local themessage = self.db.profile.warningmsg
				if self.db.profile.threatwarn and UnitExists("target") then --to avoid issues when Tricking with FoK and no target
					local isTanking,_,threatpct,_,threatvalue = UnitDetailedThreatSituation(destName, "target");
					if threatvalue == nil then threatvalue, threatpct = 0, 0 end
					local rawtotank = threatvalue * (100 / threatpct - 1)
					if not isTanking and (threatpct > self.db.profile.threatpercent or rawtotank <= self.db.profile.threatraw) then
						threatpct = tostring(mfloor(threatpct))
						themessage = themessage .. " Careful! You're at " .. threatpct.."% threat on " .. UnitName("target") .. "."
					end
				end
				if self.db.profile.targetwarn then SendChatMessage(themessage, "WHISPER", nil, destName) end
			elseif eventType == "SPELL_AURA_APPLIED" and destName == pName and self.db.profile.selfwarn then --tricks applied to you
				self:Pour(sourceName.." tricked you!")
				PlaySoundFile("Interface\\AddOns\\TrickOrTreat\\sounds\\alert.mp3")
			elseif eventType == "SPELL_AURA_REMOVED" and sourceName == pName and destName == self.last_tricked then --threat transfer expired
				self.extra_dmg = mfloor(self.dmgCounter_target - self.dmgCounter_target/damageMod)
				if self.db.profile.selfreport then self:Print(self.dmgCounter_self.." damage dealt during Tricks. Target did "..self.extra_dmg.." extra.") end
				if self.db.profile.targetreport then SendChatMessage("Tricks gave you " .. self.extra_dmg .. " extra damage.", "WHISPER", nil, self.last_tricked) end
				
				self:AddActor(self.last_tricked, self.extra_dmg)
				
				self.parse_dmg_target = false
				self.parse_dmg_self = false
				triggerEvent("TRICKS_EXPIRED")
			end
		end

		--damage accrual
		if self.parse_dmg_self then
			if (eventType == "SPELL_DAMAGE" or eventType == "SPELL_PERIODIC_DAMAGE" or eventType == "RANGE_DAMAGE") and sourceName == pName then
				self.dmgCounter_self = AddDamage(self.dmgCounter_self, arg12, arg13)
			elseif eventType == "SWING_DAMAGE" and sourceName == pName then
				self.dmgCounter_self = AddDamage(self.dmgCounter_self, arg9, arg10)
			end
		end
		if self.parse_dmg_target then
			if (eventType == "SPELL_DAMAGE" or eventType == "SPELL_PERIODIC_DAMAGE" or eventType == "RANGE_DAMAGE") and sourceName == self.last_tricked then
				self.dmgCounter_target = AddDamage(self.dmgCounter_target, arg12, arg13)
			elseif eventType == "SWING_DAMAGE" and sourceName == self.last_tricked then
				self.dmgCounter_target = AddDamage(self.dmgCounter_target, arg9, arg10)
			end
		end

	end
end

--[[ Relevant SpellIDs (dummy reference ftw)
>tricks cast on target
SPELL_CAST_SUCCESS 57934 player target => used to detect player casting on other (primer)
SPELL_AURA_APPLIED 57934 player player 

>transferring threat
SPELL_AURA_REMOVED 57934 player player => if self damage is 0 when this fires, the following events don't happen
SPELL_AURA_APPLIED 57933 player target => alert is sent to primered target, or player is alerted if a rogue tricks the player
SPELL_AURA_APPLIED 59628 player player 

expiration
SPELL_AURA_REMOVED 59628 player player 
SPELL_AURA_REMOVED 57933 player target => followup tell is sent with damage report
 	]]